//------------------------------------------------------------------------
//
// OpenDCC - DCC Packet Receiver Module
//
// Copyright (c) 2008 Alex Shepherd
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      DccRx.c
// author:    Alex Shepherd
// webpage:   http://opendcc.sourceforge.net/
// history:   2008-01-23 as Initial Version based on Wolfgang Kufer's
//                          dcc_receiver.[hc] files 
//            2008-04-14 kw added code for Sniffer
//                          This has only one capture on timer 0; so I used
//                          Timer 2 to do the system clock and PWM -> see TimerPWM.c
//                          timer was removed here.
//            2008-04-19 kw added code to evaluate the polarity
//                          (should remove memset)
//            2008-04-22 kw added test code for railcom
//
//------------------------------------------------------------------------
// credit:    This software is substantially based on work by
//            Wolfgang Kufer (kufer@gmx.de) His original project
//            can be found here: http://www.opendcc.de
//------------------------------------------------------------------------
//
// purpose:   This library module implement a DCC Packet Receiver that uses
//            an External Interrupt line INTx and a timer to decode DCC packets
//
// hardware:  using INT0 and Timer0
//
//------------------------------------------------------------------------

#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "TimerPWM.h"           // import systick_128us

#include "DccRx.h"

#define DCC_POLA_CHECK   1      // 0 = standard receive
                                // 1 = add measurement of polarity

#define DCC_RAILCOM      0      // 0 = standard
                                // 1 = add code to test railcom
                                //     (generates dummy railcom message (on command xxx)

//---------------------------------------------------------------------------
// Define TIMER0 handling macros
//
// Define the TIMER0 Prescaler bits
// #define T0_PRESCALER   8    // may be 1, 8, 64, 256, 1024
//---------------------------------------------------------------------------

#if   (T0_PRESCALER==1)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(0<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==8)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==64)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==256)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==1024)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(1<<CS00))
#else
  #error T0_PRESCALER is not defined or is not a supported value
#endif

// This are DCC timing definitions from NMRA
#define PERIOD_1   116L          // 116us for DCC 1 pulse - do not change
#define PERIOD_0   232L          // 232us for DCC 0 pulse - do not change

#define T87US (F_CPU * 77L / T0_PRESCALER / 1000000L)
// #define T_GAP (F_CPU * PERIOD_1 * 5 / 4 / T0_PRESCALER / 1000000L)  // 1.25 * PERIOD_1
#define T_GAP ((F_CPU / T0_PRESCALER * PERIOD_1 * 5 / 4 / 1000000L)%256)  // 1.25 * PERIOD_1

#if (F_CPU > 30000000)
  #warning possible long int overflow - please check all define calculations
#endif
#if (T87US > 254)
  #error T87US too big, use either larger prescaler or slower processor
#endif
#if (T87US < 32)
  #error T87US too small, use either smaller prescaler or faster processor
#endif

typedef enum
  {
    WAIT_PREAMBLE = 0,
    WAIT_RAILCOM,                   // check for a cutout
    WAIT_PREAMBLE2,
    WAIT_START_BIT,
    WAIT_DATA,
    WAIT_END_BIT
  } DccRxWaitState ;

struct
  {
    DccRxWaitState  State ;         // current state inside isr
    int8_t          LastIntTime ;   // time stamp (signed char) - domain: local count
    int8_t          LastDataReady ; // time stamp (signed char) - domain: sysclock (128us)
    uint8_t         DataReady ;     // message complete flag
    uint8_t         BitCount ;      // local counter
    uint8_t         TempByte ;      // local storage for data
    DCC_MSG         PacketBuf;      // local message buffer
    DCC_MSG         PacketCopy;     // interface message buffer
  } DccRx ;


//------------------------------------------------------------------------
// DCC Receive Routine
//
// Howto:    uses two interrupts: a rising edge in DCC polarity triggers INT0;
//           in INT0, Timer0 with a delay of 87us is started.
//           On Timer0 Overflow Match the level of DCC is evaluated and
//           parsed.
//
//                           |<-----116us----->|
//
//           DCC 1: _________XXXXXXXXX_________XXXXXXXXX_________
//                           ^-INT0
//                           |----87us--->|
//                                        ^Timer-INT: reads zero
//
//           DCC 0: _________XXXXXXXXXXXXXXXXXX__________________
//                           ^-INT0
//                           |----------->|
//                                        ^Timer-INT: reads one
//           
//           Polarity Detection:
//           after preamble, we look at the time gap for the first 0
//
//           Data:   |----1----|----1----|---------0---------|
//           Positiv: XXXXX_____XXXXX_____XXXXXXXXXX__________XXXXXXXXXX
//           Int:           ^         ^ gap=1,0  ^     gap=2         ^
//
//           Negativ: _____XXXXX_____XXXXX__________XXXXXXXXXX__________
//           Int:       ^         ^          ^  gap=1,5    ^
//
//
//           Railcom Detection:
//           There is a cutout after the end bit. the cutout is detected as = 0
//           So if we get a single 0 (duration equal four times PERIOD_1)
//           directly a the start a of a preamble -> This must be railcom
//
//           Railcom Test:
//           1. ensure correct polarity (means: invert polarity if neccessary)
//           2. dump timestamp every rising int
//           3. at endbit of message: set flag - railcom_answer
//           4. in main: check flag and correct time gap -> and send answer
//
// Result:   1. The received message is stored in DCC_MSG 'PacketCopy'
//           2. The flag DccRx.DataReady is set.
//
//
//------------------------------------------------------------------------

#if (DCC_RAILCOM == 1)
 int8_t last_dcc_edge;       // systimer: here we go last edge
 uint8_t railcom_window;     // flag: if true - main should watch for timing window to send railcom
#endif

ISR(INT0_vect) 
  {
    OCR0 = TCNT0 + T87US ;  // advance compare value.
  
    TIMSK  |= (1<<OCIE0);   // Enable Timer0 Compare Match A Interrupt
	TIFR   |= (1<<OCF0);    // Clear  Timer0 Compare Match A Flag 

    #if (DCC_RAILCOM ==1)
       last_dcc_edge = TCNT2; // save system timer to calculate the gap.
    #endif
  }


ISR(TIMER0_COMP_vect)
  {
    uint8_t DccBitVal ;

    #if (DCC_RAILCOM == 1)
        DccBitVal = (DCC_PORT_IN & (1<<DCC_BIT )) ;
        if (MCUCR & (1<<ISC00)) DccBitVal = !DccBitVal;  // if int was on rising edge - invert data
    #else
        // Read the DCC input value, if it's low then its a 1 bit, otherwise it is a 0 bit
        DccBitVal = !(DCC_PORT_IN & (1<<DCC_BIT )) ;
    #endif
  
    // Disable Timer0 Compare Match A Interrupt

    #ifdef TIMSK
        TIMSK  &= ~(1<<OCIE0);  // TIMSK  &= ~(1<<OCIE0A);  
    #elif defined(TIMSK0)
        TIMSK0 &= ~(1<<OCIE0A);
    #else
        #error: Timer0 Compare Match A Interrupt Bits not defined
    #endif

    sei();

    DccRx.BitCount++;

    switch( DccRx.State )
      {
        case WAIT_PREAMBLE:
            if( DccBitVal )
              {
                DccRx.PacketBuf.Qualifier &= ~MSG_RAILCOM;  
                DccRx.State = WAIT_PREAMBLE2 ;
              }
            else
                DccRx.State = WAIT_RAILCOM ;   // if we receive a single 0 at start of preamble
                                               // could be railcom
	        break;
      
        case WAIT_RAILCOM:
            if( DccBitVal )
              {
                int8_t gap;                                   // note: cast down to int8!
                gap = systick_128us - DccRx.LastDataReady;    // time since last data valid
                if ((gap >= 2) && (gap <= 4))  DccRx.PacketBuf.Qualifier |= MSG_RAILCOM;    
                
                DccRx.State = WAIT_PREAMBLE2 ;
              }
            else
                DccRx.BitCount = 0 ;
                DccRx.State = WAIT_PREAMBLE2 ;
	        break;
      
        case WAIT_PREAMBLE2:      // this was formerly the only inital state
            if( DccBitVal )
              {
                if( DccRx.BitCount >= 10 )
                    DccRx.State = WAIT_START_BIT ;
              }
            else
                DccRx.BitCount = 0 ;
	        break;
      
        case WAIT_START_BIT:
            if( !DccBitVal )
              {
                DccRx.State = WAIT_DATA ;
                // memset( &DccRx.PacketBuf, 0, sizeof( DCC_MSG ) );  // kw: I think this is obsolete
                DccRx.PacketBuf.Size = 0;

                // We now have 1 too many PreambleBits so decrement before copying
                DccRx.PacketBuf.PreambleBits = DccRx.BitCount - 1 ;

                DccRx.BitCount = 0 ;
                DccRx.TempByte = 0 ;
                #if (DCC_POLA_CHECK == 1)
                  {
                    // determine polarity
                    // if we have a gap of 116us, we see this 0-Bit in the first half of the cell
                    //                            and we are synced with DCC
                    // if we have a gap of 174us, we see this 0-Bit in the second half of the cell.
                    //                            and we are not synced - so we may need to invert 
                    // so we calculate the gap and compare it to the middle
                    // middle is PERIOD_1 * 1.25
                    // (all in the same modulo domain as the timer0)
                    int8_t gap;

                    gap = TCNT0 - DccRx.LastIntTime;
                    gap = gap - T_GAP;
                    if (gap >= 0)
                      {
                        #if (DCC_RAILCOM ==1)
                          if (MCUCR & (1<<ISC00)) MCUCR &= ~(1<<ISC00);    //  we need to invert the active edge of INT0.
                          else MCUCR |= (1<<ISC00);
                        #else
                          DccRx.PacketBuf.Qualifier |= MSG_POLARITY;
                        #endif
                        
                      }
                    else
                      {
                        #if (DCC_RAILCOM ==1)
                        #else
                          DccRx.PacketBuf.Qualifier &= ~MSG_POLARITY;  
                        #endif
                      }                 
                  }
                #endif
              }
            else
              {
                #if (DCC_POLA_CHECK == 1)
                    DccRx.LastIntTime = TCNT0;     // save time stamp
                #endif
              }
            break;
      
        case WAIT_DATA:
            DccRx.TempByte = ( DccRx.TempByte << 1 ) ;
            if( DccBitVal )
                DccRx.TempByte |= 1 ;

            if( DccRx.BitCount == 8 )
              {
                if( DccRx.PacketBuf.Size == MAX_DCC_MESSAGE_LEN ) // Packet is too long - abort
                  {
                    DccRx.State = WAIT_PREAMBLE ;
                    DccRx.BitCount = 0 ;
                  }
                else
                  {
                    DccRx.State = WAIT_END_BIT ;
                    DccRx.PacketBuf.Data[ DccRx.PacketBuf.Size++ ] = DccRx.TempByte ;
                  }
              }
            break;

        case WAIT_END_BIT:
            if( DccBitVal ) // End of packet?
              {
                DccRx.State = WAIT_PREAMBLE ;
                #if (DCC_RAILCOM == 1)
                    // set polarity according to active edge for interrupt
                    if (MCUCR & (1<<ISC00)) DccRx.PacketBuf.Qualifier |= MSG_POLARITY;
                    else  DccRx.PacketBuf.Qualifier &= ~MSG_POLARITY;
                    //
                    last_dcc_edge = TCNT2;  // save system timer to calculate the gap.
                                            // do it here again to reset time
                    railcom_window = 1;
                #endif
                memcpy( &DccRx.PacketCopy, &DccRx.PacketBuf, sizeof( DCC_MSG ) );
                DccRx.DataReady = 1 ;
                DccRx.LastDataReady = systick_128us;   // remember timestamp - for railcom detection
              }
            else  // Get next Byte
                DccRx.State = WAIT_DATA ;

            DccRx.BitCount = 0 ;
            DccRx.TempByte = 0 ;
	  }
  }

void initDccRx(void)
  {
    memset( &DccRx, 0, sizeof( DccRx ) );

    GICR  |= (1<<INT0);                     // Enable INT0
    MCUCR |= (1<<ISC01) | (1<<ISC00);    	//  The rising edge of INT0 generates an interrupt request.

    //       (1<<ISC01) | (0<<ISC00);    	//  The falling edge of INT0 generates an interrupt request.
                                            //  MCUCR &= ~(1<<ISC00)

    // Init the TIMER0

    TCCR0 = (0 << FOC0) | (0 << WGM00) | (0 << COM01) | (0 << COM00) | (0 << WGM01) | T0_PRESCALER_BITS ;

    DccRx.PacketBuf.Qualifier = 0;
    DccRx.DataReady = 0;

    #if (DCC_RAILCOM == 1)
        railcom_window = 0;
    #endif
  }


uint8_t getDccRxMsg( DCC_MSG * pBuf )
  {
    uint8_t returnStatus = 0 ;
  
    // We need to do this check with interrupts disabled
    cli();
    if( DccRx.DataReady )
      {
        memcpy( pBuf, &DccRx.PacketCopy, sizeof( DCC_MSG ) );
        DccRx.DataReady = 0 ;
		returnStatus = 1 ;
      }
    sei();
  
    return returnStatus ;
  }


// This is RAILCOM Channel 1 window
// typically broadcast, 2 bytes
//#define T80us (F_CPU * 80L / T2_PRESCALER / 1000000L)
#define T80us (F_CPU * 80L / 8 / 1000000L)
#define T95us (F_CPU * 95L / 8 / 1000000L)

// This is RAILCOM Channel 2 window
// typically only if our address, 2,4,6 bytes
//#define T193us (F_CPU * 193L / T2_PRESCALER / 1000000L)
#define T193us (F_CPU * 193L / 8 / 1000000L)
#define T208us (F_CPU * 208L / 8 / 1000000L)

void dummy(void)
  {
    PORTA = T80us;
  }

unsigned char railcom_test = 0;

void watch_railcom(void)
  {
    #if (DCC_RAILCOM == 1)
    
    uint8_t gap;
        if (railcom_test)
          {
            if (railcom_window == 0) return;

            gap = TCNT2 - last_dcc_edge;

            if (gap < T80us) return;
            if (gap > T95us) 
              {
                railcom_window = 0;  // window missed
                return;
              }
            else
              {
                UDR0 = 0xAA;         // dummy
                railcom_window = 0;  // window processed
              }
          }
    #endif
  }

